@@ -2,10 +2,11 @@
 import ast
 from typing import cast, Union, TYPE_CHECKING
 
-from ..utils.mixins import CommonEqualityMixin
+from nbsafety.utils import CommonEqualityMixin
 
 if TYPE_CHECKING:
-    from typing import List, Tuple, Union
+    from typing import List, Sequence, Tuple, Union
+    from nbsafety.types import SupportedIndexType
 
 
 class CallPoint(CommonEqualityMixin):
@@ -23,8 +24,9 @@ class CallPoint(CommonEqualityMixin):
 
 
 class AttrSubSymbolChain(CommonEqualityMixin):
-    def __init__(self, symbols: 'List[Union[str, CallPoint]]'):
-        self.symbols: 'Tuple[Union[str, CallPoint], ...]' = tuple(symbols)
+    def __init__(self, symbols: 'Sequence[Union[SupportedIndexType, CallPoint]]'):
+        # FIXME: each symbol should distinguish between attribute and subscript
+        self.symbols: 'Tuple[Union[SupportedIndexType, CallPoint], ...]' = tuple(symbols)
         self.call_points = tuple(filter(lambda x: isinstance(x, CallPoint), self.symbols))
 
     def __hash__(self):
@@ -36,9 +38,9 @@ class AttrSubSymbolChain(CommonEqualityMixin):
 
 class GetAttrSubSymbols(ast.NodeVisitor):
     def __init__(self):
-        self.symbol_chain: List[Union[str, CallPoint]] = []
+        self.symbol_chain: List[Union[str, int, Tuple[Union[str, int], ...], CallPoint]] = []
 
-    def __call__(self, node: 'Union[ast.Attribute, ast.Subscript, ast.Call]') -> 'AttrSubSymbolChain':
+    def __call__(self, node: 'Union[ast.Attribute, ast.Subscript, ast.Call, ast.Name]') -> 'AttrSubSymbolChain':
         self.visit(node)
         self.symbol_chain.reverse()
         return AttrSubSymbolChain(self.symbol_chain)
@@ -48,10 +50,18 @@ class GetAttrSubSymbols(ast.NodeVisitor):
             self.symbol_chain.append(CallPoint(node.func.attr))
             self.visit(node.func.value)
         elif isinstance(node.func, ast.Subscript):
-            self.symbol_chain.append(CallPoint(node.func.slice))
-            self.visit(node.func.value)
+            if isinstance(node.func.slice, ast.Constant):
+                self.symbol_chain.append(CallPoint(str(node.func.slice.value)))
+            elif isinstance(node.func.slice, ast.Index) and isinstance(node.func.slice.value, (ast.Str, ast.Num)):
+                if isinstance(node.func.slice.value, ast.Str):
+                    self.symbol_chain.append(CallPoint(node.func.slice.value.s))
+                else:
+                    self.symbol_chain.append(CallPoint(str(node.func.slice.value.n)))
+                self.visit(node.func.value)
         elif isinstance(node.func, ast.Name):
             self.symbol_chain.append(CallPoint(node.func.id))
+        elif isinstance(node.func, ast.Call):
+            pass
         else:
             raise TypeError('invalid type for node.func %s' % node.func)
 
@@ -61,12 +71,25 @@ class GetAttrSubSymbols(ast.NodeVisitor):
 
     def visit_Subscript(self, node):
         node_slice = node.slice
-        if isinstance(node_slice, ast.Index):
+        if isinstance(node_slice, ast.Constant):
+            self.symbol_chain.append(node_slice.value)
+        elif isinstance(node_slice, ast.Index):
             slice_index = node_slice.value
             if isinstance(slice_index, ast.Str):
                 self.symbol_chain.append(slice_index.s)
             elif isinstance(slice_index, ast.Num):
                 self.symbol_chain.append(slice_index.n)
+            elif isinstance(slice_index, ast.Tuple):
+                elts = []
+                for v in slice_index.elts:
+                    if isinstance(v, ast.Num):
+                        elts.append(v.n)
+                    elif isinstance(v, ast.Str):
+                        elts.append(v.s)
+                    else:
+                        break
+                else:
+                    self.symbol_chain.append(tuple(elts))
             elif isinstance(slice_index, ast.Name):
                 # FIXME: hack to make the static checker stop here
                 # In the future, it should try to attempt to resolve
@@ -84,9 +107,6 @@ class GetAttrSubSymbols(ast.NodeVisitor):
     def visit_Name(self, node):
         self.symbol_chain.append(node.id)
 
-    def visit_Str(self, node):
-        return
-
     def generic_visit(self, node):
         # raise ValueError('we should never get here: %s' % node)
         # give up
@@ -97,7 +117,8 @@ def get_attrsub_symbol_chain(maybe_node: Union[str, ast.Attribute, ast.Subscript
     if isinstance(maybe_node, (ast.Attribute, ast.Subscript, ast.Call)):
         node = maybe_node
     else:
-        node = cast(Union[ast.Attribute, ast.Subscript, ast.Call], cast(ast.Expr, ast.parse(maybe_node).body[0]).value)
+        node = cast('Union[ast.Attribute, ast.Subscript, ast.Call]',
+                    cast(ast.Expr, ast.parse(maybe_node).body[0]).value)
     if not isinstance(node, (ast.Attribute, ast.Subscript, ast.Call)):
         raise TypeError('invalid type for node %s' % node)
     return GetAttrSubSymbols()(node)
