@@ -1,13 +1,13 @@
 # -*- coding: utf-8 -*-
 import logging
+import sys
 
-import pytest
-
-from .utils import assert_bool, make_safety_fixture, skipif_known_failing
+from test.utils import assert_bool, make_safety_fixture, skipif_known_failing
 
 logging.basicConfig(level=logging.ERROR)
 
 # Reset dependency graph before each test
+# _safety_fixture, _safety_state, run_cell_ = make_safety_fixture(setup_cells=['%safety trace_messages enable'])
 _safety_fixture, _safety_state, run_cell_ = make_safety_fixture()
 
 
@@ -28,13 +28,6 @@ def assert_detected(msg=''):
     assert_bool(stale_detected(), msg=msg)
 
 
-def assert_detected_if_full_propagation(msg=''):
-    if _safety_state[0].no_stale_propagation_for_same_cell_definition:
-        assert_not_detected(msg=msg)
-    else:
-        assert_detected(msg=msg)
-
-
 def assert_false_positive(msg=''):
     """
     Same as `assert_detected` but asserts a false positive.
@@ -63,6 +56,20 @@ def test_simplest():
     assert_detected('should have detected b has stale dep on old a')
 
 
+def test_readme_example():
+    run_cell('def eval_model_1(): return 0.5')
+    run_cell('def eval_model_2(): return 0.85')
+    run_cell('models = {"model_1": eval_model_1, "model_2": eval_model_2}')
+    output = """
+best_acc, best_model = max((f(), name) for name, f in models.items())
+logging.info(f'The best model was {best_model} with an accuracy of {best_acc}.')
+    """
+    run_cell(output)
+    run_cell('def eval_model_1(): return 0.9')
+    run_cell(output)
+    assert_detected('`models` depends on stale `eval_model_1`')
+
+
 def test_subscript_dependency():
     run_cell('lst = [0, 1, 2]')
     run_cell('x = 5')
@@ -85,7 +92,6 @@ def test_long_chain():
     assert_detected('f has stale dependency on old value of a')
 
 
-@skipif_known_failing
 def test_for_loop_liveness_check():
     run_cell('x = 0')
     run_cell('y = x + 1')
@@ -305,20 +311,18 @@ def bar():
     assert_detected('Did not detect stale dependency of `accum` on `foo` and `bar`')
 
 
-@pytest.mark.parametrize("no_stale_propagation_for_same_cell_definition", [True, False])
-def test_for_loop_with_map(no_stale_propagation_for_same_cell_definition):
-    _safety_state[0].no_stale_propagation_for_same_cell_definition = no_stale_propagation_for_same_cell_definition
+def test_for_loop_with_map():
     run_cell("""
 accum = 0
 foo = [1, 2, 3, 4, 5]
 for ret in map(lambda x: x * 5, foo):
     accum += ret
-logging.info(accum)
 """)
+    run_cell('logging.info(accum)')
     assert_not_detected('no stale dep foo -> accum')
     run_cell('foo = [0]')
     run_cell('logging.info(accum)')
-    assert_detected_if_full_propagation('should detect stale dep foo -> accum unless only propagating past cell bounds')
+    assert_detected('should detect stale dep foo -> accum unless only propagating past cell bounds')
 
 
 def test_redefined_function_over_list_comp():
@@ -375,6 +379,81 @@ def bar():
     assert_detected('Did not detect stale dependency of `accum` on `foo` and `bar`')
 
 
+def test_symbol_callpoint():
+    run_cell('x = 0')
+    run_cell('y = x + 1')
+    run_cell("""
+def f():
+    return y + 7
+""")
+    run_cell('z = f()')
+    assert_not_detected()
+    run_cell('x = 42')
+    run_cell('z = f()')
+    assert_detected('`y` as referenced in call of function `f` is stale')
+
+
+def test_symbol_callpoint_2():
+    run_cell('x = 0')
+    run_cell('y = x + 1')
+    run_cell("""
+def f():
+    y = 9
+    return y + 7
+""")
+    run_cell('z = f()')
+    assert_not_detected()
+    run_cell('x = 42')
+    run_cell('z = f()')
+    assert_not_detected()
+
+
+def test_symbol_callpoint_3():
+    run_cell('x = 0')
+    run_cell('y = x + 1')
+    run_cell("""
+def f():
+    return y + 7
+""")
+    run_cell('z = f()')
+    assert_not_detected()
+    run_cell('x = 42')
+    run_cell("""
+def f():
+    return y + 7
+""")
+    run_cell('z = f()')
+    assert_detected('`y` as referenced in call of function `f` is stale')
+
+
+def test_symbol_callpoint_4():
+    run_cell('x = 0')
+    run_cell('y = x + 1')
+    run_cell("""
+def f(y):
+    return y + 7
+""")
+    run_cell('z = f(5)')
+    assert_not_detected()
+    run_cell('x = 42')
+    run_cell('z = f(5)')
+    assert_not_detected('`y` as referenced in call of function `f` is an arg and therefore not stale')
+
+
+def test_function_arg_independent_of_outer_staleness():
+    run_cell('x = 0')
+    run_cell('y = x + 1')
+    run_cell("""
+def f(y):
+    return y + 7
+""")
+    run_cell('z = f(5)')
+    assert_not_detected()
+    run_cell('x = 42')
+    run_cell('logging.info(z)')
+    assert_not_detected('`z` independent of `x`')
+
+
 def test_lambda_capture():
     run_cell('x = 42')
     run_cell('lst = []')
@@ -432,9 +511,7 @@ def test_variable_scope_2():
     assert_not_detected("Updating y should solve the problem")
 
 
-@pytest.mark.parametrize("no_stale_propagation_for_same_cell_definition", [True, False])
-def test_default_args(no_stale_propagation_for_same_cell_definition):
-    _safety_state[0].no_stale_propagation_for_same_cell_definition = no_stale_propagation_for_same_cell_definition
+def test_default_args():
     run_cell("""
 x = 7
 def foo(y=x):
@@ -445,7 +522,7 @@ def foo(y=x):
     run_cell('x = 10')
     assert_not_detected()
     run_cell('b = foo()')
-    assert_detected_if_full_propagation("Should have detected stale dependency of fn foo() on x")
+    assert_detected("Should have detected stale dependency of fn foo() on x")
 
 
 def test_same_pointer():
@@ -466,7 +543,8 @@ def test_func_assign_objs():
 a = [1]
 b = [1]
 c = [2]
-d = [3]
+d = [3]""")
+    run_cell("""
 def func(x, y=a):
     e = [c[0] + d[0]]
     f = [x[0] + y[0]]
@@ -482,18 +560,28 @@ def func(x, y=a):
     e = [c[0] + d[0]]
     f = [x[0] + y[0]]
     return f
-z = func(c)
-""")
+# z = func(c)""")
+    run_cell('z = func(c)')
     run_cell('logging.info(z[0])')
     assert_not_detected()
+    run_cell('logging.info(z)')
+    assert_not_detected()
     run_cell('c = [3]')
+    run_cell('logging.info(z)')
+    assert_detected("Should have detected stale dependency of z on c")
+    run_cell('c = [77]')
     run_cell('logging.info(z[0])')
     assert_detected("Should have detected stale dependency of z on c")
     run_cell('z = func(c)')
     run_cell('logging.info(z[0])')
     assert_not_detected()
-    run_cell('b = [4]')
-    run_cell('d = [1]')
+    run_cell('logging.info(z)')
+    assert_not_detected()
+    run_cell('b = [40]')
+    run_cell('d = [4]')
+    run_cell('logging.info(z[0])')
+    assert_not_detected("Changing b and d should not affect z")
+    run_cell('logging.info(z)')
     assert_not_detected("Changing b and d should not affect z")
 
 
@@ -533,9 +621,7 @@ z = func(c)
     assert_not_detected("Changing b and d should not affect z")
 
 
-@pytest.mark.parametrize("no_stale_propagation_for_same_cell_definition", [True, False])
-def test_func_assign_helper_func(no_stale_propagation_for_same_cell_definition):
-    _safety_state[0].no_stale_propagation_for_same_cell_definition = no_stale_propagation_for_same_cell_definition
+def test_func_assign_helper_func():
     run_cell("""
 x = 3
 a = 4
@@ -548,7 +634,7 @@ y = f()
 """)
     run_cell('x = 4')
     run_cell('logging.info(y)')
-    assert_detected_if_full_propagation("Should have detected stale dependency of y on x")
+    assert_detected("Should have detected stale dependency of y on x")
     run_cell('y = f()')
     run_cell('logging.info(y)')
     assert_not_detected()
@@ -557,9 +643,7 @@ y = f()
     assert_not_detected("Changing a should not affect y")
 
 
-@pytest.mark.parametrize("no_stale_propagation_for_same_cell_definition", [True, False])
-def test_func_assign_helper_func_2(no_stale_propagation_for_same_cell_definition):
-    _safety_state[0].no_stale_propagation_for_same_cell_definition = no_stale_propagation_for_same_cell_definition
+def test_func_assign_helper_func_2():
     run_cell("""
 x = 3
 a = 4
@@ -572,7 +656,7 @@ y = f()()
 """)
     run_cell('x = 4')
     run_cell('logging.info(y)')
-    assert_detected_if_full_propagation("Should have detected stale dependency of y on x")
+    assert_detected("Should have detected stale dependency of y on x")
 
 
 def test_branching():
@@ -609,7 +693,6 @@ def test_identity_checking():
     assert_not_detected('`y` was not mutated')
 
 
-@skipif_known_failing
 def test_identity_checking_obj():
     # To get this working properly, we need to create datasyms for all of the values in the literal
     run_cell('y = [7]')
@@ -632,6 +715,7 @@ class Foo(object):
     assert_detected('y depends on stale attrval x.x')
 
 
+@skipif_known_failing
 def test_attributes_2():
     run_cell("""
 class Foo(object):
@@ -900,7 +984,6 @@ def test_list_mutation_2():
     assert_detected('`x` depends on stale `lst`')
 
 
-@skipif_known_failing
 def test_list_mutation_from_attr():
     run_cell('s = "hello X world X how X are X you X today?"')
     run_cell('lst = []')
@@ -1091,6 +1174,7 @@ def test_single_line_dictionary_literal_fix_stale_deps():
     run_cell('logging.info(d)')
     assert_false_positive('`d`s stale dep fixed, but this is hard to detect '
                           'since we did not yet have a DataSymbol for `d[foo]` when staleness introduced')
+    # assert_not_detected('`d`s stale dep fixed')
     run_cell('foo = 8')
     run_cell('logging.info(d)')
     assert_detected('`d` depends on stale `foo`')
@@ -1156,7 +1240,6 @@ for i in [a, b, c]:
     assert_not_detected('`i` should not depend on `a` at end of for loop')
 
 
-@skipif_known_failing
 def test_for_loop_partial_dep():
     run_cell('lst = list(range(10))')
     run_cell('s = 0')
@@ -1169,7 +1252,7 @@ for i in range(5):
     assert_not_detected('`s` does not depend on last entry of `lst`')
     run_cell('lst[1] = 22')
     run_cell('logging.info(s)')
-    assert_detected('`s` does depend on second entry of `lst`')
+    assert_false_negative('`s` does depend on second entry of `lst` but tracing every iteration of loop is slow')
 
 
 def test_same_cell_redefine():
@@ -1246,7 +1329,8 @@ def test_exception_stack_unwind():
         return ';'.join([
             f'can_pass = len({safety_state}.trace_state.stack) == {size} '
             f'and len({safety_state}.attr_trace_manager.stack) == {size}',
-            f'setattr(builtins, {test_passed}, getattr(builtins, {test_passed}) and can_pass)'
+            f'setattr(builtins, {test_passed}, getattr(builtins, {test_passed}) and can_pass)',
+            f'print(len({safety_state}.trace_state.stack), "vs", {size}) '
         ])
     try:
         run_cell(f"""
@@ -1379,9 +1463,8 @@ def test_tuple_unpack_hard():
     assert_detected('`a` depends on stale `x`')
 
 
-@pytest.mark.parametrize("no_stale_propagation_for_same_cell_definition", [True, False])
-def test_attr_dep_with_top_level_overwrite(no_stale_propagation_for_same_cell_definition):
-    _safety_state[0].no_stale_propagation_for_same_cell_definition = no_stale_propagation_for_same_cell_definition
+@skipif_known_failing
+def test_attr_dep_with_top_level_overwrite():
     run_cell("""
 class Foo:
     def __init__(self):
@@ -1392,7 +1475,7 @@ foo.y = x + 7
 """)
     run_cell('x = 43')
     run_cell('logging.info(foo)')  # this should be a 'deep' usage of foo
-    assert_detected_if_full_propagation('logging.info could display `foo.y` which depends on x')
+    assert_detected('logging.info could display `foo.y` which depends on x')
     run_cell('foo.y = 70')
     assert_not_detected('we just fixed stale dep of `foo.y` by changing its deps')
     run_cell('x = foo.y + 7')
@@ -1418,6 +1501,7 @@ def test_time_line_magic():
     assert_detected('`b` has stale dep on `a`')
 
 
+@skipif_known_failing
 def test_cell_magic():
     run_cell('a = 0')
     run_cell('b = a + 1')
@@ -1476,3 +1560,204 @@ class Foo:
     run_cell('y = 42')
     run_cell('logging.info(z)')
     assert_not_detected('`z` independent of stale `y`')
+
+
+def test_no_rhs_propagation():
+    run_cell('import numpy as np')
+    run_cell('x = np.random.random(10)')
+    run_cell('y = np.random.random(10)')
+    run_cell('inds = np.argsort(x)')
+    run_cell('x = x[inds]')
+    run_cell('y = y[inds]')
+    assert_not_detected('`inds` should not be considered stale since it appears on RHS of assignment')
+
+
+def test_if_true():
+    run_cell('y = 0')
+    run_cell('z = 42')
+    run_cell("""
+if True:
+    x = y + 1
+else:
+    x = z + 1
+""")
+    run_cell('z = 43')
+    run_cell('logging.info(x)')
+    assert_not_detected('`x` not dependent on `z`')
+    run_cell('y = 99')
+    run_cell('logging.info(x)')
+    assert_detected('`x` dependent on old `y`')
+
+
+def test_tracing_reenabled_after_dup_funcall():
+    run_cell('x = 0')
+    run_cell("""
+def f():
+    return 42
+""")
+    run_cell("""
+f()
+f()
+y = x + 1
+""")
+    run_cell('x = 42')
+    run_cell('logging.info(y)')
+    assert_detected()
+
+
+def test_one_time_tracing_func():
+    run_cell('x = 0')
+    run_cell('y = 1')
+    run_cell("""
+def f(p):
+    if p:
+        return x
+    else:
+        return y
+""")
+    run_cell('z = f(False) + 1\nz = f(True) + 1')
+    run_cell('y = 2')
+    run_cell('logging.info(z)')
+    assert_not_detected()
+    run_cell('x = 3')
+    run_cell('logging.info(z)')
+    assert_false_negative('tracing should have been reactivated but this is hard')
+
+
+def test_tracing_disable_with_nested_calls():
+    run_cell('%safety trace_messages enable')
+    run_cell('y = 0')
+    run_cell("""
+def f():
+    return y
+""")
+    run_cell("""
+def g(flag):
+    if flag:
+        return f()
+    else:
+        return 2
+""")
+    run_cell("""
+g(False)
+x = g(True) + 1
+""")
+    run_cell('y = 42')
+    run_cell('logging.info(x)')
+    assert_false_negative('`x` has dep on stale `y` but capturing this is hard')
+
+
+def test_dict():
+    run_cell('d = {}; d[0] = 0')
+    run_cell('x = d[0] + 1')
+    run_cell('d = {}; d[0] = 42')
+    run_cell('logging.info(x)')
+    assert_detected('`x` has dependency on old value of `d[0]`')
+
+
+def test_default_dict():
+    run_cell('from collections import defaultdict')
+    run_cell('d = defaultdict(dict); d[0][0] = 0')
+    run_cell('x = d[0][0] + 1')
+    run_cell('d = defaultdict(dict); d[0][0] = 42')
+    run_cell('logging.info(x)')
+    assert_detected('`x` has dependency on old value of `d[0][0]`')
+
+
+def test_mutate_arg():
+    run_cell('import numpy as np')
+    run_cell('x = np.ones(5)')
+    run_cell('y = x + 1')
+    run_cell('np.random.shuffle(x)')
+    run_cell('logging.info(y)')
+    assert_detected('`y` has a dependency on an old value of `x`')
+
+
+def test_mutate_arg_special_cases():
+    run_cell('import numpy as np')
+    run_cell('x = np.random.random(10)')
+    run_cell('y = np.ones(5)')
+    run_cell('np.random.seed(42)')
+    run_cell('logging.info(x)')
+    assert_detected('`x` depends on previous seed')
+    run_cell('logging.info(y)')
+    assert_not_detected('`y` has no dependency on seed')
+    run_cell('x = y + 5')
+    run_cell('logging.info(y)')
+    assert_not_detected()
+    run_cell('logging.info(x)')
+    assert_not_detected('`logging.info` does not mutate `y`')
+
+
+def test_augassign_does_not_overwrite():
+    run_cell('x = 0')
+    run_cell('y = 1')
+    run_cell('z = x + 2')
+    run_cell('z += y')
+    run_cell('x = 42')
+    run_cell('logging.info(z)')
+    assert_detected('`z` depends on old value of `x`')
+
+
+def test_context_manager():
+    run_cell("""
+from contextlib import contextmanager
+
+@contextmanager
+def foo():
+    yield 42
+""")
+    run_cell('with foo() as bar: x = bar + 7')
+    run_cell('bar = 43')
+    run_cell('logging.info(x)')
+    assert_detected('`x` depends on old value of `bar`')
+
+
+def test_decorator():
+    run_cell('foo = lambda f: f')
+    run_cell("""
+@foo
+def bar(x):
+    return x + 42
+""")
+    run_cell('y = bar(7)')
+    run_cell('foo = lambda f: lambda x: f(x + 9)')
+    run_cell('logging.info(y)')
+    assert_detected('`y` depends on call to `bar` which has stale decorator `@foo`')
+
+
+def test_magics_dont_break_things():
+    run_cell("""
+%time dummy = 0
+x = 0
+y = x + 1
+""")
+    run_cell('x = 42')
+    run_cell('logging.info(y)')
+    assert_detected('`y` depends on old value `x`')
+
+
+if sys.version_info >= (3, 8):
+    def test_walrus_simple():
+        run_cell("""
+if (x := 1) > 0:
+    y = x + 1
+""")
+        run_cell('x = 42')
+        run_cell('logging.info(y)')
+        assert_detected('`y` depends on old value of `x`')
+
+    def test_walrus_fancy():
+        run_cell("""
+if (x := (y := (z := 1) + 1) + 1) > 0:
+    a = x + 1
+""")
+        run_cell('y = 42')
+        run_cell('logging.info(z)')
+        assert_not_detected('`z` does not depend on `y`')
+        run_cell('logging.info(y)')
+        assert_not_detected('`y` is updated')
+        run_cell('logging.info(x)')
+        assert_detected('`x` depends on old value of `y`')
+        run_cell('logging.info(a)')
+        assert_detected('`a` depends on old value of `y`')
